@@ -10,11 +10,14 @@ typedef struct {
     PyObject *callback;
     time_t seconds;
     char called;
+    PyObject *greenlet;
 } TimerObject;
 
 extern PyTypeObject TimerObjectType;
 
-TimerObject* TimerObject_new(long seconds, PyObject *callback, PyObject *args, PyObject *kwargs);
+TimerObject* TimerObject_new(long seconds, PyObject *callback, PyObject *args, PyObject *kwargs, PyObject *greenlet);
+
+void fire_timer(TimerObject *timer);
 
 int is_active_timer(TimerObject *timer);
 
